class Token():
    __slots__ = [
        'text',
        'lemma',
        'pos',
        'coarse_pos',
        'data_dic',
        'id',
        'is_stop',
        'is_alpha',
        'isnumeric',
        'islower',
        'istitle',
        'isupper'
    ]
